{
module SetPiParser where

import SetPiLexer
import qualified TypesPi as Pi


data Process = Nil
             | Par { procPar1 :: Process, procPar2 :: Process, procParPos :: AlexPosn }
             | Repl { procRepl :: Process, procReplPos :: AlexPosn }
             | New { procNewVar :: Term, procNewType :: Term, procNewCont :: Process, procNewPos :: AlexPosn }
             | In { procInCh :: Term, procInBind :: Term, procInType :: Term, procInCont :: Process, procInPos :: AlexPosn }
             | Out { procOutCh :: Term, procOutMsg :: Term, procOutCont :: Process, procOutPos :: AlexPosn }
             | Let { procLetPattern :: Term, procLetPatternType :: Term, procLetMatch :: Term,
                     procLetThen :: Process, procLetElse :: Process, procLetPos :: AlexPosn }
             | IfE { procIfExp :: SetExp, procIfThen :: Process, procIfElse :: Process, procIfPos :: AlexPosn }
             | SetE { procSetExp :: SetExp, procEnterCont :: Process, procEnterPos :: AlexPosn }
             | SelectE { procSelVar :: Term, procSelType :: Term, procSelExp :: Maybe SetExp, procSelCont :: Process, procSelPos :: AlexPosn }
             | Event { procEventName :: Set, procEventTerm :: Term, procEventType :: EventType, procEventCont :: Process, procEventPos :: AlexPosn }
             | Lock { procLockSet :: [Set], procLockCont :: Process, procLockPos :: AlexPosn }
             | Unlock { procUnlockSet :: [Set], procUnlockCont :: Process, procUnlockPos :: AlexPosn }
             | MacroCall { procMacroName :: String, procMacroArgs :: [Term], procMacroPos :: AlexPosn }
               deriving (Eq)

data EventType = NonInjective | Injective
               deriving (Eq, Show)

instance Show Process where
    show (Nil) = ""
    show (Par p1 p2 pos) = "(" ++ (indent $ show p1) ++ ") | (" ++ (indent $ show p2) ++ ")"
    show (Repl p pos) = "!\n" ++ show p
    show (New x t p pos) = "new " ++ show x ++ " : " ++ show t ++ ";\n" ++ show p
    show (In ch x t p pos) = "in(" ++ show ch ++ ", " ++ show x ++ ":" ++ show t ++ ");\n" ++ show p
    show (Out ch msg p pos) = "out(" ++ show ch ++ ", " ++ show msg ++ ");\n" ++ show p
    show (Let t1 ty1 t2 p1 p2 pos) = "let " ++ show t1 ++ ": " ++ show ty1 ++ " = " ++ show t2 ++ " in\n"
                                 ++ (indent $ show p1) ++ "else\n" ++ (indent $ show p2)
    show (IfE b p1 p2 pos) = "if " ++ show b ++ " then\n" ++ (indent $ show p1) ++ " else\n" ++ (indent $ show p2)
    show (SetE b p pos) = "set " ++ show b ++ ";\n" ++ show p
    show (SelectE x t b p pos) = "select " ++ show x ++ " : " ++ show t ++
                                 (case b of Just b -> " where " ++ show b; Nothing -> "") ++ " in " ++ show p
    show (Event e x ty p pos) = (if ty == Injective then "inj-event " else "event ") ++
                                show e ++ "(" ++ show x ++ ");\n" ++ show p
    show (Lock s p pos) = "lock(" ++ showSep "," s ++ ");\n" ++ show p
    show (Unlock s p pos) = "unlock(" ++ showSep "," s ++ ");\n" ++ show p

data SetExp = SetIn Term Set
            | And SetExp SetExp
            | Or SetExp SetExp
            | Not SetExp
              deriving (Eq)

instance Show SetExp where
    show (SetIn t s) = show t ++ " in " ++ show s
    show (And b1 b2) = "(" ++ show b1 ++ " and " ++ show b2 ++ ")"
    show (Or b1 b2) = "(" ++ show b1 ++ " or " ++ show b2 ++ ")"
    show (Not b) = "not " ++ show b

data Decl =
  FuncDecl { funcDeclArgs :: [Term], funcDeclVisible :: Visibility, funcDeclPos :: AlexPosn }
  | SetDecl { setDeclType :: Term, setDeclSetType :: SetDeclType, setDeclPos :: AlexPosn }
  | MacroDef { macroArgs :: [String], macroProc :: Process }
  | NameDef { nameType :: Term, nameVisible :: Visibility }
  | TypeDecl { typeDef :: Term }
  deriving (Eq)

instance Show Decl where
    show (FuncDecl name args private pos) = "fun " ++ name ++ "/" ++ show (List.length args) ++ "."
    show (Reduc name args res private pos) = "reduc " ++ name ++ "(" ++ showSep "," args ++ ") = " ++ show res

data SetDeclType = SetDeclT | EventDeclT | InjEventDeclT
                   deriving (Eq, Show)

data Set = Set { setName :: String }
           deriving (Eq, Ord)
instance Show Set where
    show (Set n) = n

data Query = Query { queryPred :: Pred, queryState :: Maybe SetExp, queryGamma :: Gamma }
  deriving (Eq, Show)
data Reduc = Reduc { reducArgs :: [Term], reducRes :: Term, reducVis :: Visibility, reducPos :: AlexPosn }
  deriving (Eq, Show)

data System = System { systemProcess :: Process,
                       systemReducs :: [Reduc],
                       systemQueries :: [Query],
                       systemDecs :: [(String, Decl)],
                       systemParams :: [Param]}
              deriving (Eq, Show)

data Visibility = Public | Private
  deriving (Eq, Show)

data Param = PVParam { pvParamName :: String, pvParamValue :: String }
             deriving(Eq)
instance Show Param where
    show (PVParam k v) = "param " ++ k ++ " = " ++ v ++ ".\n"

warning :: String -> AlexPosn -> a -> a
warning msg pos = trace (sourceName pos ++ ":" ++ show (sourceLine pos) ++
                         ":" ++ show (sourceColumn pos) ++ ": " ++ msg)

}

%name parser
%tokentype {Token}

%token ident    {T (Ident $$) _}
       'agree'  {T Agree _}
       'fun'	{T Fun _}
       'pred'	{T Pred _}
       'reduc'	{T Reduc _}
       'set'	{T Set _}
       'forall' {T ForAll _}
       'in' 	{T In _}
       'out' 	{T Out _}
       'let' 	{T Let _}
       'if'	{T If _}
       'lock' 	{T Lock _}
       'unlock' {T Unlock _}
       'inj-agree' {T InjAgree _}
       'inj-event' {T InjEvent _}
       'then' 	{T Then _}
       'else' 	{T Else _}
       'event' 	{T Event _}
       'new' 	{T New _}
       'and'	{T And _}
       'or'	{T Or _}
       'not' 	{T Not _}
       'private' {T Private _}
       'update'	{T Update _}
       'query'  {T Query _}
       'process' {T Process _}
       'select' {T Select _}
       'where' 	{T Where _}
       '==>' 	{T Implies _}
       '<>'	{T Neq _}
       '='	{T Eq _}
       '('	{T OpenPar _}
       ')'	{T ClosePar _}
       '{' 	{T OpenBrace _}
       '}'	{T CloseBrace _}
       '['	{T OpenBracket _}
       ']'	{T CloseBracket _}
       ';'	{T SemiColon _}
       ':' 	{T Colon _}
       ','	{T Comma _}
       '.'	{T Dot _}
       '!'	{T Bang _}
       '|'	{T Par _}

%%

visibility
: '[' 'private' ']' { Pi.Private }
| { Pi.Public };

term :: {Pi.Term}
: var {$1} | func {$1} | tuple {$1};

args :: {[Pi.Term]}
: term {[$1]}
| term ',' args {($1 : $3)};

var :: {Pi.Term}
: ident { Pi.Var $1 };

func :: {Pi.Term}
: ident '(' args ')' { Pi.Func Pi.TFunc $1 $3 };

tuple :: {Pi.Term}
: '(' args ')' { Pi.Func Pi.TTuple $1 $3 };


pred :: {Pi.Pred}
: ident '(' args ')' { Pi.Pred $1 $3 };


pattern :: {(Pi.Term, Pi.Term)}
: tuplePattern { $1 } | varBind { $1 } | match { $1 };

patterns
: pattern {(\ (p, t) -> ([p], [t])) $1}
| pattern ',' patterns {(\ (p, t) (ps, ts) -> (p : ps, t : ts)) $1 $3};

tuplePattern
: '(' patterns ')' {(\ (ps, ts) -> (Pi.mkTuple ps, Pi.mkTuple ts)) $2};

varBind
: var ':' term {($1, $3)};

match
: '=' term ':' term {($1, $3)};

set
: ident { Pi.Set $1 };

sets
: set { [$1] }
| set ',' sets { $1 : $3 };

boolExp
: boolExp1 'or' boolExp { Pi.Or $1 $3 };

boolExp1
: boolExp2 'and' boolExp1 { Pi.And $1 $3 };

boolExp2
: term 'in' set { Pi.SetIn $1 $3 }
| term 'not' 'in' set { Pi.Not (Pi.SetIn $1 $4) }
| 'not' boolExp2 { Pi.Not $2 }
| '(' boolExp ')' { $2 };

process :: {Pi.Process}
: 'in' '(' term ',' pattern ')' cont {(\ (p, t) -> Pi.In $3 p t $7) $5 }
| 'out' '(' term ',' term ')' cont { Pi.Out $3 $5 $7 }
| '!' process {Pi.Repl $2}
| 'new' var ':' var cont { Pi.New $2 $4 $5 }
| 'let' pattern '=' term 'in' process else {(\ (p t) -> Pi.Let p t $4 $6 $7) $1}
| 'if' boolExp 'then' process else { Pi.If $2 $4 $5 }
| 'lock' '(' sets ')' cont { Pi.Lock $3 $5 }
| 'unlock' '(' sets ')' cont { Pi.Unlock $3 $5 }
| 'event' ident '(' args ')' cont { Pi.Event (Pi.Set $2) $4 NonInjective $6 }
| 'inj-event' ident '(' args ')' cont { Pi.Event (Pi.Set $2) $4 Injective $6 }
| 'update' boolExp cont { Pi.SetE $2 $3 }
| 'select' var ':' term 'where' boolExp cont { Pi.SelectE $2 $4 (Just $6) $7}
| ident '(' args ')' { Pi.Macro $1 $3 }
| process '|' process { Pi.Par $1 $3 };

cont :: {Pi.Process}
: ';' process { $2 }
| { Pi.Nil };

else :: {Pi.Process}
: 'else' process {$2}
| { Pi.Nil };


system :: {Pi.System}
: 'process' process { Pi.System $2 [] [] [] [] }
| 'fun' ident '(' args ')' visibility '.' system
  { $8 { systemDecs = decl : systemDecs $8 }
    where decl = ($2, Pi.FuncDecl $4 $6) }
| 'set' ident ':' term '.' system
  { $6 { systemDecs = decl : systemDecs $6 }
    where decl = ($2, Pi.SetDecl $2 $4) }
| 'let' ident '=' process '.' system
  { $6 { systemDecs = decl : systemDecs $6 }
    where decl = ($2, Pi.MacroDef $2 $4) }
;
